@@ -31,7 +31,14 @@
 
 #include <ev.h>
 
-#include <openssl/ssl.h>
+#include "ssl_compat.h"
+
+#ifdef NGHTTP2_OPENSSL_IS_WOLFSSL
+#  include <wolfssl/options.h>
+#  include <wolfssl/openssl/ssl.h>
+#else // !NGHTTP2_OPENSSL_IS_WOLFSSL
+#  include <openssl/ssl.h>
+#endif // !NGHTTP2_OPENSSL_IS_WOLFSSL
 
 #include "shrpx_rate_limit.h"
 #include "shrpx_connection.h"
@@ -52,6 +59,7 @@ class Worker;
 class Downstream;
 struct WorkerStat;
 struct DownstreamAddrGroup;
+struct SharedDownstreamAddr;
 struct DownstreamAddr;
 #ifdef ENABLE_HTTP3
 class Http3Upstream;
@@ -59,14 +67,17 @@ class Http3Upstream;
 
 class ClientHandler {
 public:
-  ClientHandler(Worker *worker, int fd, SSL *ssl, const StringRef &ipaddr,
-                const StringRef &port, int family, const UpstreamAddr *faddr);
+  ClientHandler(Worker *worker, int fd, SSL *ssl,
+                const std::string_view &ipaddr, const std::string_view &port,
+                int family, const UpstreamAddr *faddr);
   ~ClientHandler();
 
   int noop();
   // Performs clear text I/O
   int read_clear();
   int write_clear();
+  // Specialized for PROXY-protocol use; peek data from socket.
+  int proxy_protocol_peek_clear();
   // Performs TLS handshake
   int tls_handshake();
   // Performs TLS I/O
@@ -96,7 +107,7 @@ public:
   void reset_upstream_write_timeout(ev_tstamp t);
 
   int validate_next_proto();
-  const StringRef &get_ipaddr() const;
+  const std::string_view &get_ipaddr() const;
   bool get_should_close_after_write() const;
   void set_should_close_after_write(bool f);
   Upstream *get_upstream();
@@ -122,7 +133,7 @@ public:
   int perform_http2_upgrade(HttpsUpstream *http);
   bool get_http2_upgrade_allowed() const;
   // Returns upstream scheme, either "http" or "https"
-  StringRef get_upstream_scheme() const;
+  std::string_view get_upstream_scheme() const;
   void start_immediate_shutdown();
 
   // Writes upstream accesslog using |downstream|.  The |downstream|
@@ -132,7 +143,7 @@ public:
   Worker *get_worker() const;
 
   // Initializes forwarded_for_.
-  void init_forwarded_for(int family, const StringRef &ipaddr);
+  void init_forwarded_for(int family, const std::string_view &ipaddr);
 
   using ReadBuf = DefaultMemchunkBuffer;
 
@@ -149,16 +160,17 @@ public:
 #ifdef ENABLE_HTTP3
   void setup_http3_upstream(std::unique_ptr<Http3Upstream> &&upstream);
   int read_quic(const UpstreamAddr *faddr, const Address &remote_addr,
-                const Address &local_addr, const uint8_t *data, size_t datalen);
+                const Address &local_addr, const ngtcp2_pkt_info &pi,
+                std::span<const uint8_t> data);
   int write_quic();
 #endif // ENABLE_HTTP3
 
   // Returns string suitable for use in "by" parameter of Forwarded
   // header field.
-  StringRef get_forwarded_by() const;
+  std::string_view get_forwarded_by() const;
   // Returns string suitable for use in "for" parameter of Forwarded
   // header field.
-  StringRef get_forwarded_for() const;
+  std::string_view get_forwarded_for() const;
 
   Http2Session *
   get_http2_session(const std::shared_ptr<DownstreamAddrGroup> &group,
@@ -167,7 +179,11 @@ public:
   // Returns an affinity cookie value for |downstream|.  |cookie_name|
   // is used to inspect cookie header field in request header fields.
   uint32_t get_affinity_cookie(Downstream *downstream,
-                               const StringRef &cookie_name);
+                               const std::string_view &cookie_name);
+
+  DownstreamAddr *get_downstream_addr_strict_affinity(
+    int &err, const std::shared_ptr<SharedDownstreamAddr> &shared_addr,
+    Downstream *downstream);
 
   const UpstreamAddr *get_upstream_addr() const;
 
@@ -178,15 +194,19 @@ public:
 
   // Stores |sni| which is TLS SNI extension value client sent in this
   // connection.
-  void set_tls_sni(const StringRef &sni);
+  void set_tls_sni(const std::string_view &sni);
   // Returns TLS SNI extension value client sent in this connection.
-  StringRef get_tls_sni() const;
+  std::string_view get_tls_sni() const;
 
   // Returns ALPN negotiated in this connection.
-  StringRef get_alpn() const;
+  std::string_view get_alpn() const;
 
   BlockAllocator &get_block_allocator();
 
+  void set_alpn_from_conn();
+
+  void set_local_hostport(const sockaddr *addr, socklen_t addrlen);
+
 private:
   // Allocator to allocate memory for connection-wide objects.  Make
   // sure that the allocations must be bounded, and not proportional
@@ -198,15 +218,20 @@ private:
   std::unique_ptr<Upstream> upstream_;
   // IP address of client.  If UNIX domain socket is used, this is
   // "localhost".
-  StringRef ipaddr_;
-  StringRef port_;
+  std::string_view ipaddr_;
+  std::string_view port_;
   // The ALPN identifier negotiated for this connection.
-  StringRef alpn_;
+  std::string_view alpn_;
   // The client address used in "for" parameter of Forwarded header
   // field.
-  StringRef forwarded_for_;
+  std::string_view forwarded_for_;
   // lowercased TLS SNI which client sent.
-  StringRef sni_;
+  std::string_view sni_;
+  // The host and port of local address where the connection is
+  // accepted.  For QUIC connection, the local address may change due
+  // to client address migration, but this value stays the same for
+  // now.
+  std::string_view local_hostport_;
   std::function<int(ClientHandler &)> read_, write_;
   std::function<int(ClientHandler &)> on_read_, on_write_;
   // Address of frontend listening socket
